#!/usr/bin/perl
use strict;
use warnings;
use File::Copy;

use make_program_input;
use init_params;
use write_params;
use run_mpi_prog;
use incr;
use validate;
use psf_parse;
use write_header;
use update_res;
use get_info;
use select;
use get_ann_lo_hi;

my $time_start = time();
my $startdir = `pwd`;
chomp($startdir);

# Get and validate command line arguments
my ($ncpu, $nodefile, $param_file) = @ARGV;
validate_commandline_args($ncpu, $nodefile, $param_file);

# Initialize PO2R, P3DR, PCUT, PSF, PPFT, and control hashes
my (%auto_param, %po2r_param, %p3dr_param,
    %pcut_param, %psf_param, %ppft_param);
my @ptlefiles = ();
my @selectfiles = ();
init_params($param_file, \%auto_param, \%po2r_param,
	    \%p3dr_param, \%pcut_param, \%psf_param,
	    \%ppft_param, \@ptlefiles);

# Open outfile and write header information
open(OUT, ">$auto_param{outfile}");
write_header($ncpu, $nodefile, \%auto_param, 
	     \%p3dr_param, \%po2r_param, 
	     \%pcut_param, \%psf_param, \%ppft_param);

# Open summary output file
my $outfile_stripped = $auto_param{outfile};
$outfile_stripped =~ s/.*\///;
my $summary_outfile = "summary_${outfile_stripped}";
open(SUM, ">$summary_outfile");

# Change to the run directory where particle (and optionally map data)
# reside and where we want the results files (new particle
# orientations and maps) to be written
chdir($auto_param{rundir});

# Create a directory where intermediate files can be moved
my $intfiles_dir = "INTFILES_${outfile_stripped}";
mkdir $intfiles_dir;

# Get pixel_size from first ptlefile and update PSF pixel_size
my $pixel_size = get_pixel_size($ptlefiles[0]);
$psf_param{pixel_size} = $pixel_size;

# Extract boxrad (packRadius in PIF standard) from image file
# Make undefined if value equal to zero
my $pif_name = get_pif_name($ptlefiles[0]);
my $boxrad_from_pif = get_boxrad($pif_name);
if($boxrad_from_pif == 0) {
    undef $boxrad_from_pif;
}

# The following code is used to allow for the specification of the box
# radius in either the PIF file or the auto3dem parameter
# file. Necessary because some PIF files are missing the box radius

# Case 1 - box radius not set in either location
unless(defined $auto_param{boxrad} || defined $boxrad_from_pif) {
    print "Box radius not specified in PIF file or parameter file\n";
    print "Either fix $pif_name or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $auto_param{boxrad} && defined $boxrad_from_pif) {
    print "Box radius specified in both PIF file and parameter file\n";
    if($auto_param{boxrad} == $boxrad_from_pif) {
	print "Box radius ($boxrad_from_pif) specified consistently\n\n";
    } else {
	print OUT "\nWarning boxrad from PIF file ($boxrad_from_pif) ";
	print OUT "not equal to boxrad from parameter file ";
	print OUT "($auto_param{boxrad})\n";
	print OUT "Value in parameter file overides PIF file\n\n";
    }
}

# Case 3 - box radius specified only in PIF file
if(defined $boxrad_from_pif && !defined $auto_param{boxrad}) {
    print OUT "Box radius from PIF file = $boxrad_from_pif\n\n";
    $auto_param{boxrad} = $boxrad_from_pif;
}

# Case 4 - box radius specified only in parameter file
if(!defined $boxrad_from_pif && defined $auto_param{boxrad}) {
    print OUT "Box radius from parameter file = $auto_param{boxrad}\n\n";
}

print OUT "Pixel size from $ptlefiles[0] = $pixel_size\n\n";
print OUT "\n";

# Use boxrad to set program parameters
unless(defined $ppft_param{pftrad_hi}) {
    $ppft_param{pftrad_hi} = $auto_param{boxrad};
}

# Set the number of CPUs to be used by each program
# ($PROG_param{max_cpu} is the upper limit)
my ($p3dr_cpu, $po2r_cpu, $pcut_cpu, $psf_cpu, $ppft_cpu);
$p3dr_cpu  = ($p3dr_param{max_cpu} < $ncpu) ? $p3dr_param{max_cpu} : $ncpu;
$po2r_cpu  = ($po2r_param{max_cpu} < $ncpu) ? $po2r_param{max_cpu} : $ncpu;
$pcut_cpu  = ($pcut_param{max_cpu} < $ncpu) ? $pcut_param{max_cpu} : $ncpu;
$psf_cpu   = ($psf_param{max_cpu}  < $ncpu) ? $psf_param{max_cpu}  : $ncpu;
$ppft_cpu  = ($ppft_param{max_cpu} < $ncpu) ? $ppft_param{max_cpu} : $ncpu;

# Start iterations one cycle earlier if we need to construct map
my $iter_start = 0;
if($auto_param{have_map}) {
    $iter_start = 1;
}

# Set things up for hollowing map
my %hollow_param;
if($auto_param{hollow_map}) {
    $hollow_param{out_file_name} = "hollow_map";
    $hollow_param{in_rad}        = $auto_param{hollow_in_rad};
    $hollow_param{out_rad}       = $auto_param{hollow_out_rad};
    $hollow_param{cut_step}      = $auto_param{hollow_cut_step}; 
    $hollow_param{cut_weight}    = $auto_param{hollow_cut_weight}; 
}

# LOOP OVER ITERATIONS
for (my $i=$iter_start; $i<=$auto_param{niter}; $i++) { 

    print OUT "*********** Iteration $i ***********\n\n";

    my $input_file;
    my $output_file;
    
    # Run PO2R or PPFT to update particle origins and orientations
    unless($i == 0) {
	printf OUT "*** Updating origins and orientations\n\n";

	if($auto_param{mode} eq 'refine') {
	    $input_file  = "po2r_input_iter_${i}";
	    $output_file = "po2r_output_iter_${i}";
	    if(($auto_param{have_map} && $i == 1) ||
	       $auto_param{new_ptles}) {
		$po2r_param{in_3d_name} = $auto_param{start_map};
	    } else {
		my $indx = $i - 1;
		$po2r_param{in_3d_name} = "map_iter_${indx}";
	    }
	    unless(-r $po2r_param{in_3d_name}) {
		printf OUT "PO2R unable to open $po2r_param{in_3d_name}\n";
		printf OUT "Exiting AUTO3DEM\n";
		exit;
	    }

	    if($auto_param{hollow_map}) {
		$hollow_param{in_3d_name} = $po2r_param{in_3d_name};
		$po2r_param{in_3d_name} = $hollow_param{out_file_name};
		make_program_input('hollow_in', 'pcut', \%hollow_param);
		run_mpi_prog($pcut_param{bin}, 'hollow_in',
				$pcut_cpu, 'hollow_out', $nodefile);
	    }

	    make_program_input($input_file, 'po2r', 
			       \%po2r_param, \@ptlefiles);
	    run_mpi_prog($po2r_param{bin},  $input_file,       
			 $po2r_cpu, $output_file, $nodefile);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	} else {
	    if(($auto_param{have_map} && $i == 1) ||
	       $auto_param{new_ptles}) {
		$ppft_param{model_filename} = $auto_param{start_map};
	    } else {
		my $indx = $i - 1;
		$ppft_param{model_filename} = "map_iter_${indx}";
	    }
	    unless(-r $ppft_param{model_filename}) {
		printf OUT "PPFT unable to open $ppft_param{model_filename}\n";
		printf OUT "Exiting AUTO3DEM\n";
		exit;
	    }

	    if($auto_param{hollow_map}) {
		$hollow_param{in_3d_name} = $ppft_param{model_filename};
		$po2r_param{model_filename} = $hollow_param{out_file_name};
		make_program_input('hollow_in', 'pcut', \%hollow_param);
		run_mpi_prog($pcut_param{bin}, 'hollow_in',
				$pcut_cpu, 'hollow_out', $nodefile);
	    }

	    $input_file  = "ppft_input_iter_${i}";
	    $output_file = "ppft_output_iter_${i}";
	    $ppft_param{pftrads_filename} = "ppft_iter_${i}.rads";
	    $ppft_param{pftres1_filename} = "ppft_iter_${i}.res1";
	    $ppft_param{pftres2_filename} = "ppft_iter_${i}.res2";
	    if(-e $ppft_param{pftrads_filename}) {
		unlink $ppft_param{pftrads_filename};
	    }
	    if(-e $ppft_param{pftres1_filename}) {
		unlink $ppft_param{pftres1_filename};
	    }
	    if(-e $ppft_param{pftres2_filename}) {
		unlink $ppft_param{pftres2_filename};
	    }
	    make_program_input($input_file, 'ppft', 
			       \%ppft_param, \@ptlefiles);
	    run_mpi_prog($ppft_param{bin},  $input_file,       
			 $ppft_cpu, $output_file, $nodefile);

	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	    move ($ppft_param{pftrads_filename}, 
		  "$intfiles_dir/$ppft_param{pftrads_filename}");
	    move ($ppft_param{pftres1_filename}, 
		  "$intfiles_dir/$ppft_param{pftres1_filename}");
	    move ($ppft_param{pftres2_filename}, 
		  "$intfiles_dir/$ppft_param{pftres2_filename}");
	    
	    # Parse PPFT output to update inner and outer radii of the
	    # annulus. Use these to set (annulus_low, annulus high)
	    # for next run of PPFT and (in_rad, out_rad) for PCUT
	    unless($auto_param{freeze_annulus}) {
		my ($ann_lo, $ann_hi) = 
		    get_ann_lo_hi($ppft_param{pftrads_filename},
				  $ppft_param{pftrad_lo},
				  $ppft_param{pftrad_step},
				  $auto_param{boxrad});
		$ppft_param{annulus_low} = $ann_lo;
		$ppft_param{annulus_high} = $ann_hi;
		$pcut_param{in_rad} = $ann_lo;
		$pcut_param{out_rad} = $ann_hi;
		print OUT "   PPFT annulus_low changed to  $ann_lo\n";
		print OUT "   PPFT annulus_high changed to $ann_hi\n";
		print OUT "   PCUT in_rad changed to       $ann_lo\n";
		print OUT "   PCUT out_rad changed to      $ann_hi\n";
	    }
	}

	# Increment the particle origin and orientation files
	incr_ptlefiles(\@ptlefiles);
    }

    print OUT "*** Evaluating selection criteria\n\n";
    
    my $res_locut_best = 1000;
    my $res_hicut_best = 1000;
    my $res_flag_best = 0;
    my $selection_best = 0;

    # Make a copy of P3DR hash for creating even and odd maps and set
    # the zero_fill to be one.
    my %p3dr_param_oddeven = %p3dr_param;
    $p3dr_param_oddeven{zero_fill} = 1.0;
    
    # LOOP OVER SELECTION CRITERIA

    # If auto_param{estimate_res} is false, then NO resolution
    # estimation will be performed and selection criterion 0 will be
    # used to construct the map

    my @jlist;
    if($auto_param{estimate_res}) {
	my $nso = $auto_param{nselect_offset};
	@jlist = (-$nso .. $nso);
    } else {
	$selection_best = 0;
	@jlist = ();
    }

    foreach my $j (@jlist) {
	
	my $suffix = "${i}_${j}";
	$suffix =~ s/-/n/;
	
	# Generate the filtered particle files
	@selectfiles = gen_filtered_ptles(\@ptlefiles, \%auto_param, $j);
	
	# Run P3DR to generate even map
	$input_file  = "p3dr_even_input_iter_$suffix";
	$output_file = "p3dr_even_output_iter_$suffix";
	$p3dr_param_oddeven{oddeven} = 2;
	$p3dr_param_oddeven{out_file_name} = "map_even";
	make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   \@selectfiles);
	run_mpi_prog($p3dr_param_oddeven{bin}, $input_file, 
		     $p3dr_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run P3DR to generate odd map
	$input_file  = "p3dr_odd_input_iter_$suffix";
	$output_file = "p3dr_odd_output_iter_$suffix";
	$p3dr_param_oddeven{oddeven} = 1;
	$p3dr_param_oddeven{out_file_name}  = "map_odd";
	make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   \@selectfiles);
	run_mpi_prog($p3dr_param_oddeven{bin}, $input_file,  
		     $p3dr_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	foreach my $selectfile (@selectfiles) {
	    move $selectfile, "$intfiles_dir/$selectfile";
	}

	# Run PCUT to cut the even map
	$input_file  = "pcut_even_input_iter_$suffix";
	$output_file = "pcut_even_output_iter_$suffix";
	$pcut_param{in_3d_name}    = "map_even";
	$pcut_param{out_file_name} = "map_even_cut";
	make_program_input($input_file, 'pcut', \%pcut_param);
	run_mpi_prog($pcut_param{bin}, $input_file, 
		     $pcut_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run PCUT to cut the odd map
	$input_file  = "pcut_odd_input_iter_$suffix";
	$output_file = "pcut_odd_output_iter_$suffix";
	$pcut_param{in_3d_name}     = "map_odd";
	$pcut_param{out_file_name}  = "map_odd_cut";
	make_program_input($input_file, 'pcut', \%pcut_param);
	run_mpi_prog($pcut_param{bin}, $input_file,  
		     $pcut_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run PSF to generate FSC curve
	$input_file  = "psf_input_iter_$suffix";
	$output_file = "psf_output_iter_$suffix";
	$psf_param{fileeven} = "map_even_cut";
	$psf_param{fileodd}  = "map_odd_cut";
	make_program_input($input_file, 'psf', \%psf_param);
	run_mpi_prog($psf_param{bin},  $input_file,       
		     $psf_cpu, $output_file, $nodefile);
	
	# Parse the PSF output file to determine current resolution and
	# high resolution to be used in next iteration
	my ($res_hicut, $res_locut, $res_flag) = 
	    psf_parse($auto_param{fsc_hicut}, $auto_param{fsc_locut},
		      $output_file);
	
	# Determine whether this selection criterion results in
	# the best resolution
	if($res_hicut < $res_hicut_best) {
	    $res_hicut_best = $res_hicut;
	    $res_locut_best = $res_locut;
	    $selection_best = $j;
	    $res_flag_best = $res_flag;
	    copy($output_file, "FSC_curve_iter_$i");
	}
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Cleanup temporary map files
	unlink 'map_even', 'map_even_cut', 'map_odd', 'map_odd_cut';
    }

    print OUT "*** Construct map using selection criterion ";
    print OUT "$selection_best\n";

    # Adjust resolutions based on the best $res_locut
    if($auto_param{estimate_res} && !$auto_param{freeze_res}) {
	update_res(\%po2r_param, \%p3dr_param, \%psf_param, \%ppft_param,
		   \%auto_param, $res_locut_best, $res_flag_best);
    }
    
    # If we have new particles that we are refining relative to an
    # existing higher resolution map, we do not need to construct a
    # new map.
    unless($auto_param{new_ptles}) {
	# Generate the filtered particle files using best selection criterion
	@selectfiles = gen_filtered_ptles(\@ptlefiles, \%auto_param,
					  $selection_best);
    
	# Run P3DR to generate full map using filtered particle files
	$p3dr_param{out_file_name} = "map_iter_${i}";
	$input_file  = "p3dr_input_iter_${i}";
	$output_file = "p3dr_output_iter_${i}";
	$p3dr_param{oddeven} = 0;
	make_program_input($input_file, 'p3dr', \%p3dr_param, \@selectfiles);
	run_mpi_prog($p3dr_param{bin}, $input_file,      
		     $p3dr_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	foreach my $selectfile (@selectfiles) {
	    move $selectfile, "$intfiles_dir/$selectfile";
	}
    }

    # Write out restart file
    my $restart_file = "restart_file_${i}";
    $restart_file = "${startdir}/${restart_file}";
    write_params($restart_file, \%auto_param, \%po2r_param,
		 \%p3dr_param, \%pcut_param, \%psf_param,
		 \%ppft_param, \@ptlefiles);

    # Write out summary information
    print SUM "*** Iteration $i\n";
    print SUM "  Estimated resolution $res_hicut_best\n";
    if($res_flag_best) {
	print SUM "  NOTE - FSC never drops below $auto_param{fsc_locut}";
	print SUM ", real resolution may be better\n";
    }
    print SUM "  Mode = $auto_param{mode}\n";
    if($auto_param{mode} eq "refine") {
	print SUM "  PO2R angle step size $po2r_param{dangle}\n";
    } else {
	print SUM "  PPFT angle step size $ppft_param{delta_theta}\n";
    }
    print SUM "  Map calculated to $p3dr_param{res_min}";
    print SUM " with Gaussian falloff to $p3dr_param{res_max}\n";
    print SUM "\n";

}

# Rename the last map generated by P3DR to be the final map
unless($auto_param{new_ptles}) {
    move $p3dr_param{out_file_name}, $auto_param{final_map};
}

# Delete intermediate map files if flag set
if($auto_param{delete_maps}) {
    unlink <map_iter_*>;
}

# Update names of starting map, final map, and output file, then write
# out continuation file

# Starting map for next run is final map from this run
$auto_param{start_map} = $auto_param{final_map};

# If the output file is of the form string{n}, increment the name of
# the next output file to string{n+1}. E.g. output2 --> output3
if($auto_param{outfile} =~ /.*\d+$/) {
    my $suffix = $auto_param{outfile};
    $suffix =~ s/.*?(\d+)$/$1/;
    $suffix++;
    $auto_param{outfile} =~ s/\d+$/$suffix/;
}

# Set the name of new final map to newmap.pif. Up to user to manually
# edit continuation file for next map.
$auto_param{final_map} = "newmap.pif";

# Add comment to alert user to check these parameters
$auto_param{start_map} .= "  # Check this";
$auto_param{final_map} .= "  # Check this";
$auto_param{outfile}   .= "  # Check this";

my $continue_file = "continue_file";
$continue_file = "${startdir}/${continue_file}";
write_params($continue_file, \%auto_param, \%po2r_param,
	     \%p3dr_param, \%pcut_param, \%psf_param,
	     \%ppft_param, \@ptlefiles);

my $time_stop = time();
my $time_elapsed = $time_stop - $time_start;
print OUT "Auto3dem finished\n";
print OUT "   Total run time = $time_elapsed seconds\n";

close(OUT);
close(SUM);
