      subroutine intlz_arrays(hkl_min, hkl_max, min_xyz, max_xyz,
     +      min_y, max_y, min_z, max_z, max_views, num_symm,
     +      first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +      max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +      first_u, last_u, max_u_slab, first_view, last_view,
     +      first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +      max_node_eqns, npix_row,
     +      iam_continues_cmpt_intrps,
     +      iam_continues_solve_fftsynth_1, iam_continues_fftsynth_2)
C                                                           INTLZ_ARRAYS
C **********************************************************************
C *                                                               *
C * Subroutine: intlz_arrays                                      *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: aamain -- The main routine.                          *
C * Function: Initializing arrays recording the partitioning      *
C *           information of data structures when running in      *
C *           multiple nodes.                                     *
C * Usage:                                                        *
C *   call intlz_arrays( hkl_min, hkl_max, min_xyz, max_xyz,      *
C *          min_y, max_y, min_z, max_z, max_views, num_symm,     *
C *          first_hk, last_hk, max_hk_slab, first_iq, last_iq,   *
C *          max_iq_slab, first_rr_out, last_rr_out,              *
C *          max_rr_out_slab, first_u, last_u, max_u_slab,        *
C *          first_view, last_view, first_y, last_y, max_y_slab,  *
C *          first_z, last_z, max_z_slab, max_node_eqns )         *
C * Input:                                                        *
C *   hkl_min (Integer) -- The minimum index of HKL space.        *
C *   hkl_max (Integer) -- The maximum index of HKL space.        *
C *   min_y, max_y, min_z, max_z (Integer) -- The minimum and     *
C *     maximum indexes of Y and Z of the the density map         *
C *   max_views (Integer) -- The maximum number of views          *
C *   num_symm (Integer) -- The number of symmetry                *
C * Output:                                                       *
C *   first_view (Integer) -- The starting index of the views     *
C *     assigned to the current node for processing.              *
C *   last_view (Integer) -- The ending index of the views        *
C *     assigned to the current node for processing.              *
C *   max_node_eqns (Integer) -- The maximum number of views      *
C *                              assigned to a node               *
C *   first_iq (Integer) -- The starting index of "iq" in this    *
C *                         node.                                 *
C *   last_iq (Integer) -- The ending index of "iq" in this node. *
C *   max_iq_slab (Integer) -- The maximum length of the segments *
C *     of the coordinates of the HKL space divided among nodes.  *
C *   min_xyz, max_xyz (Integer) -- The minimum and maximum       *
C *     indexes of X, Y, Z of the the density map                 *
C *   first_y, last_y (Integer) -- The first and last index of Y  *
C *                                in this node                   *
C *   max_y_slab (Integer) -- The maximum width of Y slab         *
C *                        the real space for this node.          *
C *   first_z (Integer) -- The starting index of coordinate Z of  *
C *   last_z (Integer) -- The ending index of coordinate Z of the *
C *                       real space for this node.               *
C *   max_z_slab (Integer) -- The maximum length of segments in   *
C *     the coordinates Z in real space divided among nodes.      *
C *   first_hk (Integer) -- The starting index of coordinate H or *
C *     K of reciprocal space for this node.                      *
C *   last_hk (Integer) -- The ending index of coordinate H or K  *
C *                        of reciprocal space for this node.     *
C *   max_hk_slab (Integer) -- The maximum length of segments in  *
C *     the coordinates H and K divided among nodes.              *
C *   first_rr_out, last_rr_out (Integer) -- The first and last   *
C *     index of RR in this node                                  *
C *   max_rr_out_slab (Integer) -- The maximum width of RR slab   *
C *                                                               *
C * Algorithm:                                                    *
C *   Fill FRST_LST_*** array:  FRST_LST_***(1, NODE) and         *
C *   FRST_LST_IQ(2, NODE) give the first and last index in the   *
C *   `***-slab' processed by node NODE.                          *
C *                                                               *
C *   If there are more nodes than values of ***.  Assign one     *
C *   *** to each  node; for excess nodes, make difference        *
C *   between first and last FRST_LST_*** indices negative.       *
C *                                                               *
C *   If there are at least as many nodes as there are values of  *
C *   ***, fill ***_LAST with the increment of ***'s from one     *
C *   node to the next. Distribute the remaining IQ values to the *
C *   first few nodes.                                            *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 0.6, Jul.6, 1998                      *
C *       Modifying the partitioning of HKs and making it         *
C *       dependent to the partitioning of IQs. The reason is:    *
C *       distribution of the result of Fourier synthesis is      *
C *       determined by the distribution of Fourier coefficients, *
C *       because synthesis is done in parallel and each node     *
C *       processes the local data independently. Former method   *
C *       partitioning HKs equally in multiple nodes is a design  *
C *       flaw and makes the result of program incorrect.         *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 7, 1999                     *
C *       Adding indexes RR and Y.                                *
C *     - R.E.Lynch 8/2000:
C *       New output:                                                  *
C *         the following set here instead of in aamain                *
C *           iam_continues_cmpt_intrps                                *
C *           iam_continues_solve_fftsynth_1                           *
C *           iam_continues_fftsynth_2                                 *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *         
C *                                                               *
C *****************************************************************
C
C-----------------------------------------------------------------------
C Initialize some index arrays, `first' and `last' are the indices
C used this node, and `slab' is the (-first+last+1).  The values of
C `first' and `last' for all nodes are stored in an array in the common
C / COM_SMALL_ARRAYS / in include file SMALL_ARRAYS.inc
C for all node
C sizes

C  Input:  -------------------------------------------------------------
C   HKL_MIN, HKL_MAX  INTEGER*4; first and last Fourier Transform
C                        index H, K, or L

C   MIN_XYZ,MAX_XYZ   INTEGER*4; minimum of min_x, min_y, min_z; maximum
C                        of max_x, max_y, max_z.

C   MIN_Y,MAX_Y, MIN_Z,MAX_Z  INTEGER*4; first and last values of Y and
C                        Z coordinates
 
C   MAX_VIEWS    INTEGER*4; maximum number of views in data-input-file.
 
C   NUM_SYMM     INTEGER*$; maximum number of symmetry operators.

C  Output:  ------------------------------------------------------------
C   The arrays below are in COMMON / COM_SMALL_ARRAYS / in the
C   include file SMALL_ARRAYS.inc

C   FRST_LST_HK( 2, 0:num_nodes-1)  INTEGER*4; first and last value of
C       H or K frame processed by each node during synthesis.
C   FIRST_HK, LAST_HK   INTEGER*4; FRST_LST_HK(1,IAM) and
C       FRST_LST_HK(2,IAM): the first and last H or K processed by this
C       node during synthesis.
C   MAX_HK_SLAB         INTEGER*4; the maximum number of H or K procesed
C       by any node.

C   FRST_LST_IQ( 2, 0:num_nodes-1)  INTEGER*4; first and last IQ
C       values used by each node.  For a single node:
C         0,1,2, ..., hkl_max
C   FIRST_IQ, LAST_IQ   INTEGER*4; FRST_LST_IQ(1,IAM) and
C       FRST_LST_IQ(2,IAM): the first and last IQ used by this node.
C   MAX_IQ_SLAB         INTEGER*4; total number of IQ planes:
C       -FRST_LST_IQ(1,0)+FRST_LST_IQ(1,NUM_NODES_1)

C Beginning of change 10/19/98 -----------------------------------------
C   FRST_LST_RR_OUT( 2, 0:num_nodes-1)  INTEGER*4;
C       first and last array index for storing a Fouier coefficient;
C       the coefficients of  cos( m * .) are A_m, m = 0, ..., HKL_MAX
C       with array indices 1,2, 4, 6, ..., 2*HKL_MAX;
C       the coefficients of  sin( m * .) are B_m, m = 1, ..., HKL_MAX;
C       with array indices     3, 5, 7, ..., 2*HKL_MAX+1;
C         M = 1,    2,   3,  ..., 2*HKL_MAX, 2*HKL_MAX+1
C             A_0, A_1, B_1, ..., A_HKL_MAX, B_HKL_MAX
C       These are obtained from the values in FRST_LST_IQ.
C End of chage 10/19/98 ------------------------------------------------
C   FIRST_RR_OUT, LAST_RR_OUT  INTEGER*4; FRST_LST_RR_OUT(1,IAM) and
C       FRST_LST_RR_OUT(2,IAM); the first and last RR used by this node.
C   MAX_RR_OUT_SLAB     INTEGER*4; total number of RR planes:
C       -FRST_LST_RR_OUT(1,0)+FRST_LST_RR_OUT(1,NUM_NODES_1)

C   FRST_LST_U( 2, 0:num_nodes-1)   INTEGER*4; first and last value of
C       U processed by each node during synthesis of H or K.
C   FIRST_U, LAST_U     INTEGER*4; FRST_LST_U(1,IAM) and
C       FRST_LST_U(2,IAM): the first and last U processed by this
C       node during synthesis of H or K.
C   MAX_U_SLAB          INTEGER*4; the maximum number of U procesed
C       by any node during synthesis of H or K.

C   FRST_LST_Y( 2, 0:num_nodes-1)   INTEGER*4; first and last value of
C       Y processed by each node during synthesis of H or K.
C   FIRST_Y, LAST_Y     INTEGER*4; FRST_LST_Y(1,IAM) and
C       FRST_LST_Y(2,IAM): the first and last Y processed by this
C       node during synthesis of H or K.
C   MAX_Y_SLAB          INTEGER*4; the maximum number of Y procesed
C       by any node during synthesis of H or K.

C   FRST_LST_VIEW( 2, 0:num_nodes-1)  INTEGER*4; first and last pixel
C       frame processed by each node.
C   FIRST_VIEW, LAST_VIEW   INTEGER*4; FRST_LST_VIEW(1,IAM) and
C       FRST_LST_IQ(2,IAM): the first and last view processed by this
C       node.

C   FRST_LST_Z( 2, 0:num_nodes-1)   INTEGER*4; first and last value of
C       Z processed by each node during synthesis of H or K.
C   FIRST_Z, LAST_Z     INTEGER*4; FRST_LST_Z(1,IAM) and
C       FRST_LST_Z(2,IAM): the first and last Z processed by this
C       node during synthesis of H or K.
C   MAX_Z_SLAB          INTEGER*4; the maximum number of Z procesed
C       by any node during synthesis of H or K.

C   MAX_NODE_EQNS       INTEGER*4; the maximum number of equations
C       which can be solved during SOLVESYS.f (by some node). 
C   NUM_INTRP_EQNS      INTEGER*4; the maximum number of equations
C       which can be solved by this node.
C-----------------------------------------------------------------------
C
      implicit none
      include '../include/allprogs.inc'
C
C Arguments
C
      integer hkl_min, hkl_max, min_xyz, max_xyz, min_y, max_y,
     +        min_z, max_z, max_views, num_symm, first_hk, last_hk,
     +        max_hk_slab, first_iq, last_iq, max_iq_slab, first_rr_out,
     +        last_rr_out, max_rr_out_slab, first_u, last_u, max_u_slab,
     +        first_view, last_view, first_y, last_y, max_y_slab,
     +        first_z, last_z, max_z_slab, max_node_eqns, npix_row

      logical iam_continues_cmpt_intrps,
     +        iam_continues_solve_fftsynth_1, iam_continues_fftsynth_2
C
C Local
C
      integer node
      logical debug
      debug = .false. 
C
C ----------------------------------------------------------------------
C     FRST_LST_HK( 1, NODE), FRST_LST( 2, NODE) give the first and last
C     H or K processed by node NODE.
C      call set_indices( num_nodes, 1, (-hkl_min+hkl_max+1), frst_lst_hk,
C     +   max_hk_slab )
C
C     FRST_LST_IQ( 1, NODE), FRST_LST_IQ( 2, NODE) give the first and
C     last Fourier Transform subscipt, H, K, or L, in the `Q-slab'
C     processed by node NODE.
C
      call set_indices( num_nodes, 0, hkl_max, frst_lst_iq, max_iq_slab)
C
C     FRST_LST_RR_OUT( 1, NODE), FRST_LST_RR_OUT( 2, NODE) give first
C     and last array indices for storing the Fourier coefficients,
C     A_M and B_M,  A_M coefficient of COS( M * . ) and B_M the 
C     coefficient of  SIN( M * . ).  B_0 is zero and is not stored.
C     For one node the indices are:  1, 2, 3, 4, 5, ...
C     for A_0, A_1, B_1, A_2, B_2, ....  The values stored in
C     FRST_LST_RR_OUT are obtained from those in FRST_LST_IQ
C
      max_rr_out_slab = 0
      do 500 node = 0, num_nodes - 1
        frst_lst_rr_out(1,node) = 2 * frst_lst_iq(1,node)
        frst_lst_rr_out(2,node) = 2 * frst_lst_iq(2,node) + 1
C
Cji add this to make to be 2*N, the same size as max_x-min_x+1= 2*hkl_max
        if (node .eq. 0) then
           frst_lst_rr_out(1,0) = 1
        endif
        if (node .eq. num_nodes - 1) then
           frst_lst_rr_out(2,node) = frst_lst_rr_out(2,node) -1
        endif

        max_rr_out_slab = max(max_rr_out_slab,
     +    -frst_lst_rr_out(1,node)+frst_lst_rr_out(2,node)+1)
  500 continue
C
Cji frst_lst_hk and max_hk_slab not used now
      do node = 0, num_nodes-1
        frst_lst_hk(1,node) = frst_lst_rr_out(1,node) 
        frst_lst_hk(2,node) = frst_lst_rr_out(2,node)
      enddo
      max_hk_slab = max_rr_out_slab
C
Cji I have changed it to be the first and last npix_row processed by node NODE
C     frst_lst_y( 1, NODE), frst_lst_y( 2, NODE ) give the first
C     and the last pixel-frame processed by node NODE
C
      call set_indices( num_nodes, 1, npix_row, frst_lst_y,
     +                  max_y_slab )
C
C     FRST_LST_U( 1, NODE), FRST_LST_U( 2, NODE) give the first and
C     last U coordinate in the slab processed by node NODE for Fourier
C     coefficients a spatial and a Fourier index.
C
      call set_indices( num_nodes, min_xyz, max_xyz, frst_lst_u,
     +                  max_u_slab )
C
C     FRST_LST_Z( 1, NODE), FRST_LST_Z( 2, NODE) give the first and
C     last Z coordinates in the `Z-slab' processed by node NODE.
C
      call set_indices( num_nodes, min_z, max_z, frst_lst_z,
     +                  max_z_slab )
C 
C     Store values used on this node
C
      first_hk       = frst_lst_hk(1,iam)
      last_hk        = frst_lst_hk(2,iam)
      first_iq       = frst_lst_iq(1,iam)
      last_iq        = frst_lst_iq(2,iam)
      first_rr_out   = frst_lst_rr_out(1,iam)
      last_rr_out    = frst_lst_rr_out(2,iam)
      first_view     = 1
      last_view      = max_views
Cji each processor deal with view_perread views and pixels in the read,comput loop 
Cji             1 <= view_perread <= num_views <= max_views

      first_u        = frst_lst_u(1,iam)
      last_u         = frst_lst_u(2,iam)
      first_y        = frst_lst_y(1,iam)
      last_y         = frst_lst_y(2,iam)
      first_z        = frst_lst_z(1,iam)
      last_z         = frst_lst_z(2,iam)
 
      if (first_view .le. last_view) then
        iam_continues_cmpt_intrps = .true. 
      else
        iam_continues_cmpt_intrps = .false.
      endif
      if (first_iq .le. last_iq) then
        iam_continues_solve_fftsynth_1 = .true. 
      else
        iam_continues_solve_fftsynth_1 = .false.
      endif
      if (first_u .le. last_u) then
        iam_continues_fftsynth_2 = .true. 
      else
        iam_continues_fftsynth_2 = .false.
      endif
 
      if( debug ) write(stdoutput,8000) iam,
     +   first_hk, last_hk, first_iq, last_iq,
     +   first_rr_out, last_rr_out, first_view, last_view,
     +   first_u, last_u, first_y, last_y,
     +   first_z, last_z,
     +   iam_continues_cmpt_intrps,
     +   iam_continues_solve_fftsynth_1, iam_continues_fftsynth_2
 8000     format(/
     +    ' DEBUG of intlz_arrays.f, iam   = ',i8/
     +    ' first_hk                       = ',i8/,
     +    ' last_hk                        = ',i8/,
     +    ' first_iq                       = ',i8/,
     +    ' last_iq                        = ',i8/,
     +    ' first_rr_out                   = ',i8/,
     +    ' last_rr_out                    = ',i8/,
     +    ' first_view                     = ',i8/,
     +    ' last_view                      = ',i8/,
     +    ' first_u                        = ',i8/,
     +    ' last_u                         = ',i8/,
     +    ' first_y                        = ',i8/,
     +    ' last_y                         = ',i8/,
     +    ' first_z                        = ',i8/,
     +    ' last_z                         = ',i8/,
     +    ' iam_continues_cmpt_intrps      = ',L8/,
     +    ' iam_continues_solve_fftsynth_1 = ',L8/,
     +    ' iam_continues_fftsynth_2       = ',L8/)
      return
      end
